#ifdef WIN32
#define _CRT_SECURE_NO_WARNINGS
#endif

// #define DEBUG_DETECT

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "object.h"
#include "util.h"
#include "file.h"
#include "location.h"
#include "code.h"
#include "function.h"
#include "source.h"

extern int	g_hypertext_cnt;
extern int	g_top_clones;
extern int	g_top_cols;
extern int	g_report_c;
extern int	g_report_assembler;
extern int	g_verbose;
extern char	*g_html_dirP;
extern int	g_ignore_macros;
extern int	g_min_match_sequence;
extern int	g_min_function_sequence;
extern int	g_mismatch_cost;
extern int	g_same_cost;
extern int	g_match_lines;
extern int  g_comparisons;
extern int	g_use;

static	Csource			*source1P;		// Source 1st side looking at
static	Csource			*source2P;		// Source 2nd side looking at
static	Cfunction		*function1P;	// Function 1st side looking at
static	Cfunction		*function2P;	// Function 2nd side looking at
static	codeT			*function_code1P;// Start of function1 code
static	codeT			*function_code2P;// Start of function2 code
static	codeT			*function_end1P;// Beyond end of function1 code
static	codeT			*function_end2P;// Beyond end of function2 code
static	codeT			*start_code1P;	// Comparing from 1st side
static	codeT			*start_code2P;	// Comparing from 2nd side
static	codeT			*end_code1P;	// May not compare to here on 1st side
static	codeT			*end_code2P;	// May not compare to here on 2nd side


extern FILE		*stdoutF;
extern FILE		*redirectF;

static Csource	*last_sourceP = 0;
static FILE		*sourceF      = 0;
static FILE		*cloneF      = 0;

typedef struct {
	unsigned int	weight;
	unsigned int	html;
} topT;

static topT		*top_clonesP       = 0;
static topT		*top_clones_endP   = 0;
static int		top_clones_cnt     = 0;

void
initialise_detect(void)
{
	int	need;

	if (g_top_clones > 0) {
		need = g_top_clones * sizeof(topT);
		top_clonesP = (topT *) xmalloc(need);
		memset(top_clonesP, 0, need);
		top_clones_endP = (topT *) (((char *) top_clonesP) + need);
	}
}

void
start_report(void)
{
	if (g_html_dirP) {
		fputs("<p>\n<table>\n", redirectF);
}	}

static int	g_lth1, g_lth2;

static void
report_clone(void)
{
	extern	FILE			*taF;
	extern	FILE			*tempF;

	static	int				lineno1, lineno2, next_lineno1, next_lineno2;
	static	Cfile			*file1P, *file2P;
	static	const Cfile		*sourcefile1P, *sourcefile2P;
	static	int				different1, different2;

	static	char	*source_name1P, *source_name2P;
	static	codeT	*at_code1P, *at_code2P;
	static	int		offset1, offset2;
	static	unsigned int	hypertext_cnt;

	static	unsigned int heap_index;
	static	topT	*heapP, *leftP, *rightP;
	static	unsigned int	weight;

	sourcefile1P  = source1P->m_fileP;
	sourcefile2P  = source2P->m_fileP;
	source_name1P = sourcefile1P->m_nameP;
	source_name2P = sourcefile2P->m_nameP;

	// Increment the detail page number
	g_hypertext_cnt += 1;

	if (taF) {
		Cclone *clone1P, *clone2P;

		clone1P = function1P->locate_clone(start_code1P, end_code1P);
		clone2P = function2P->locate_clone(start_code2P, end_code2P);

		fprintf(taF, "R C%p C%p\n", clone1P, clone2P);
		if (g_html_dirP && (g_report_c || g_report_assembler)) {
			fprintf(tempF, "(R C%p C%p) { html=%u }\n", clone1P, clone2P, g_hypertext_cnt);
	}	}

	if (top_clonesP) {
		// Use a heap with the smallest weight at the top

		weight = g_lth1 + g_lth2;
		if (top_clonesP->weight < weight) {
			++top_clones_cnt;
			heap_index = 1;
			heapP      = top_clonesP;
			// Add entry to the heap
			for (;;) {
				leftP = heapP + heap_index;
				if (top_clones_endP <= leftP) {
					goto found_heap_entry;
				}
				rightP = leftP + 1;
				if ((top_clones_endP <= rightP) || (leftP->weight <= rightP->weight)) {
					if (weight <= leftP->weight) {
						goto found_heap_entry;
					}
					*heapP       = *leftP;
					heapP        = leftP;
					heap_index <<= 1;
					continue;
				}
				if (weight <= rightP->weight) {
found_heap_entry:	heapP->weight = weight;
					heapP->html   = g_hypertext_cnt;
					break;
				}
				*heapP       = *rightP;
				heapP        = rightP;
				heap_index   = (heap_index << 1) + 1;
			}
	}	}

	if (g_html_dirP) {
		if (last_sourceP != source1P) {
			if (sourceF) {
				fprintf(sourceF,"</table>\n<p>\nEnd of report</body>\n</html>\n");
				fclose(sourceF);
			}
			hypertext_cnt = source1P->m_hypertext_cnt;
			sprintf(g_filename,"%s/%u.html", g_html_dirP, hypertext_cnt);
			sourceF = open_file(g_filename);
			last_sourceP = source1P;

			stdoutF       = sourceF;

			fprintf(stdoutF, HTML_HEADER "<title>Clones in %s", source_name1P);
			fprintf(stdoutF, "</title>\n<body>\n<a href=index.html>Top</a>\n<h3>Clones in %s", source_name1P);
			fprintf(stdoutF, "</h3>\n<table>\n");
			
			// On top level page write
			stdoutF = redirectF;
			fprintf(stdoutF, "<tr><td><a href=\"%u.html\">%s", hypertext_cnt, source_name1P);
			fprintf(stdoutF, "</a></td></tr>\n");
		}

		cloneF = 0;
		if (g_report_c || g_report_assembler) {
			sprintf(g_filename,"%s/%u.html", g_html_dirP, g_hypertext_cnt);
			cloneF = open_file(g_filename);
			fputs(
HTML_HEADER
"<script language=\"javascript\">\n"
"<!--\n"
"var a, s;\n"
"var a_shown='none'\n"
"var s_shown='none'\n"
"function toggleA() {\n"
"  for (i = 0; i < a.length; i++) { a[i].style.display = a_shown}\n"
"  a_shown = (a_shown == 'none' ? 'inline' : 'none')\n"
"}\n"
"function toggleS() {\n"
"  for (i = 0; i < s.length; i++) { s[i].style.display = s_shown}\n"
"  s_shown = (s_shown == 'none' ? 'inline' : 'none')\n"
"}\n"
"function init(){ \n"
"  var ai = 0;\n"
"  var si = 0;\n"
"  var tr = document.all.tags(\"tr\");\n"
"  var e, className, i;\n"
"  for (i = 0; i < tr.length; i++) {\n"
"    e = tr[i];\n"
"    className = \"\" + e.className;\n"
"    switch (className) {\n"
"      case 'a': ++ai; break;\n"
"      case 's': ++si; break;\n"
"  } }\n"
"  a  = new Array(ai);\n"
"  s  = new Array(si);\n"
"  ai = 0;\n"
"  si = 0;\n"
"  for (i = 0; i < tr.length; i++) {\n"
"    e = tr[i];\n"
"    className = \"\" + e.className;\n"
"    switch(className) {\n"
"    case 'a': a[ai] = e; ++ai; break;\n"
"    case 's': s[si] = e; ++si; break;\n"
"} } }\n"
"//-->\n"
"</script>\n"
"</head>\n"
"<body onLoad=\"init()\">\n"
"<p>\n"
"[<a href=index.html>Top</a>] [<a href=\"javascript:toggleA()\">Toggle</a>] Assembler\n"
"[<a href=\"javascript:toggleS()\">Toggle</a>] Source\n"
"<p>\n"
				  , cloneF);
		}
		
		stdoutF = sourceF;
		fputs("<tr><td><i>", stdoutF);
		if (cloneF) {
			fprintf(stdoutF, "<a href=\"%u.html\">%s", g_hypertext_cnt, function1P->m_nameP);
		}
		if (cloneF) {
			fputs("</a>", stdoutF);
		}
		fprintf(stdoutF, "</i> <font color=\"red\">[%d]</font></td>\n    <td><i>%s.%s", (int) (end_code1P - start_code1P), source_name2P, function2P->m_nameP);
		fprintf(stdoutF, "</i> <font color=\"blue\">[%d]</font></td>\n", (int) (end_code2P - start_code2P));
		fputs("</i></td>\n</tr>\n", stdoutF);
		if (!cloneF) {
			stdoutF = redirectF;
			return;
		}
		stdoutF = cloneF;
	} 
	fputs("<p>\n<table>\n", stdoutF);
	fprintf(stdoutF, "<tr><td><b><a href=%u.html>%s</a>", source1P->m_hypertext_cnt, source_name1P);
	fprintf(stdoutF, "</b></td>\n    <td><b><a href=%u.html>%s</a>", source2P->m_hypertext_cnt, source_name2P);
	fputs("</b></td>\n</tr>\n", stdoutF);
	fprintf(stdoutF, "<tr><td><i>%s", function1P->m_nameP);
	fprintf(stdoutF, "</i> <font color=\"red\">[%d]</font></td>\n    <td><i>%s", (int) (end_code1P - start_code1P), function2P->m_nameP);
	fprintf(stdoutF, "</i> <font color=\"blue\">[%d]</font></td>\n</tr>\n", (int) (end_code2P - start_code2P));

	if (g_report_c || g_report_assembler) {

		fputs("<tr><td><hr></hr></td><td><hr></hr></td></tr>", stdoutF); 

		at_code1P = start_code1P;
		at_code2P = start_code2P;

		if (g_report_c) {

			if (at_code1P < end_code1P) {
				next_lineno1 = at_code1P->m_source_lineno;
				file1P       = at_code1P->m_fileP;
			} else {
				next_lineno1 = -1;
				file1P       = 0;
			}
			if (at_code2P < end_code2P) {
				next_lineno2 = at_code2P->m_source_lineno;
				file2P       = at_code2P->m_fileP;
			} else {
				next_lineno2 = -1;
				file2P       = 0;
			}
			different1   = 0;
			different2   = 0;
		}

		offset1 = at_code1P - function_code1P;
		offset2 = at_code2P - function_code2P;
		for (; at_code1P < end_code1P || at_code2P < end_code2P; ) {
			
			if (g_report_assembler) {
				if (end_code1P <= at_code1P || (at_code1P->m_matchP && at_code1P->m_matchP != at_code2P) ) {
					fprintf(stdoutF, "<tr class=a><td></td>\n    <td><font color=\"blue\">%d: ", offset2);
					put_html(at_code2P->m_codeP, stdoutF);
					if (at_code2P->m_jump != JUMP_UNDEFINED) {
						fprintf(stdoutF, " -> %d", offset2 + at_code2P->m_jump);
					}
					++at_code2P;
					++offset2;
					fputs("</font></td>\n</tr>\n", stdoutF);
					different2 = 1;
				} else if (!at_code1P->m_matchP) {
					fprintf(stdoutF, "<tr class=a><td><font color=\"red\">%d: ", offset1);
					put_html(at_code1P->m_codeP, stdoutF);
					if (at_code1P->m_jump != JUMP_UNDEFINED) {
						fprintf(stdoutF, " -> %d", offset1 + at_code1P->m_jump);
					}
					++at_code1P;
					++offset1;
					fputs("</font></td>\n    <td>", stdoutF);
					different1 = 1;
				
					if (at_code2P < end_code2P && !at_code2P->m_matchP) {
						fprintf(stdoutF, "<font color=\"blue\">%d: ", offset2);
						put_html(at_code2P->m_codeP, stdoutF);
						if (at_code2P->m_jump != JUMP_UNDEFINED) {
							fprintf(stdoutF, " -> %d", offset2 + at_code2P->m_jump);
						}
						++at_code2P;
						++offset2;
						fputs("</font>", stdoutF);
						different2 = 1;
					}
					fputs("</td>\n</tr>\n", stdoutF);
				} else {
					fprintf(stdoutF, "<tr class=a");
					if (at_code1P->m_fileP == at_code2P->m_fileP && at_code1P->m_source_lineno == at_code2P->m_source_lineno) {
						fprintf(stdoutF, " bgColor=#D0D0D0");
					}
					fprintf(stdoutF, "><td>%d: ", offset1);
					put_html(at_code1P->m_codeP, stdoutF);
					if (at_code1P->m_jump != JUMP_UNDEFINED) {
						fprintf(stdoutF, " -> %d", offset1 + at_code1P->m_jump);
					}
					++at_code1P;
					++offset1;
					fprintf(stdoutF, "</td>\n    <td>%d: ", offset2);
					put_html(at_code2P->m_codeP, stdoutF);
					if (at_code2P->m_jump != JUMP_UNDEFINED) {
						fprintf(stdoutF, " -> %d", offset2 + at_code2P->m_jump);
					}
					++at_code2P;
					++offset2;
					fputs("</td>\n</tr>\n", stdoutF);
				} 
			} else {
				// Not reporting assembler

				if (end_code1P <= at_code1P || (at_code1P->m_matchP && at_code1P->m_matchP != at_code2P) ) {
					++at_code2P;
					different2 = 1;
				} else if (!at_code1P->m_matchP) {
					++at_code1P;
					different1 = 1;
				
					if (at_code2P < end_code2P && !at_code2P->m_matchP) {
						++at_code2P;
					}
				} else {
					++at_code1P;
					++at_code2P;
			}	} 

			if (g_report_c) {

				lineno1      = next_lineno1;
				lineno2      = next_lineno2;
				next_lineno1 = ((at_code1P < end_code1P) ? at_code1P->m_source_lineno : -1);
				next_lineno2 = ((at_code2P < end_code2P) ? at_code2P->m_source_lineno : -1);

				if (lineno1 == next_lineno1) {
					if (!file2P || lineno2 == next_lineno2) {
						continue;
					}
					fputs("<tr class=s><td></td>\n    <td><font color=\"blue\">", stdoutF);
					file2P->showline(&sourcefile2P, lineno2);
					fputs("</font></td>\n</tr>\n", stdoutF);
				} else if (lineno2 == next_lineno2) {
					if (!file1P || lineno1 == next_lineno1) {
						continue;
					}
					fputs("<tr class=s><td><font color=\"red\">", stdoutF);
					file1P->showline(&sourcefile1P, lineno1);
					fputs("</font></td>\n    <td></td></tr>\n", stdoutF);
				} else if (file1P || file2P) {
					fputs("<tr class=s", stdoutF);
					if (!different1 && !different2 && sourcefile1P == sourcefile2P && lineno1 == lineno2) {
						fputs(" bgColor=#D0D0D0", stdoutF);
					}
					fputs("><td>", stdoutF);
					if (file1P) {
						if (!different1) {
							file1P->showline(&sourcefile1P, lineno1);
						} else {
							fputs("<font color=\"red\">", stdoutF);
							file1P->showline(&sourcefile1P, lineno1);
							fputs("</font>", stdoutF);
					}	}
					fputs("</td><td>", stdoutF);
					if (file2P) {
						if (!different2) {
							file2P->showline(&sourcefile2P, lineno2);
						} else {
							fputs("<font color=\"blue\">", stdoutF);
							file2P->showline(&sourcefile2P, lineno2);
							fputs("</font>", stdoutF);
					}	}
					fputs("</td></tr>\n", stdoutF);
				}
				file1P = ((at_code1P < end_code1P) ? at_code1P->m_fileP : 0);
				file2P = ((at_code2P < end_code2P) ? at_code2P->m_fileP : 0);
				different1 = 0;
				different2 = 0;
			}
		}
	}
	fputs("</table>\n", stdoutF);
	if (g_html_dirP) {
		fputs("<p>End of report\n</body>\n</html>\n", stdoutF);
		fclose(stdoutF);
		stdoutF = redirectF;
	}
}

void
finalize_report(void)
{
	static	topT	*topP, *atP, *endP, *heapP, *leftP, *rightP;
	static	int		heap_index, cnt, i;
	static	FILE	*topF;

	if (sourceF) {
		fprintf(sourceF,"</table>\n<p>\n<p>End of report</body>\n</html>\n");
		fclose(sourceF);
	}
	
	if (top_clonesP) {
		cnt = top_clones_cnt;
		if (cnt > g_top_clones) {
			cnt = g_top_clones;
		}
		if (cnt) {
			topP = atP = (topT *) xmalloc(cnt *sizeof(topT));
			for (endP = atP + cnt; ; ) {
				if (top_clonesP->html) {
					*atP++     = *top_clonesP;
					if (endP <= atP) {
						break;
				}	}
				heapP      = top_clonesP;
				heap_index = 1;
				for (;;) {
					leftP = heapP + heap_index;
					if (top_clones_endP <= leftP) {
						heapP->weight = 0xFFFFFFFF;
						break;
					}
					rightP = leftP + 1;
					if (top_clones_endP <= rightP || leftP->weight <= rightP->weight) {
						*heapP       = *leftP;
						heapP        = leftP;
						heap_index <<= 1;
						continue;
					}
					*heapP     = *rightP;
					heapP      = rightP;
					heap_index = (heap_index << 1) + 1;
			}	}
			sprintf(g_filename,"%s/top.html", g_html_dirP);
			topF = open_file(g_filename);
			fprintf(topF, HTML_HEADER "<title>Top %d largest clones</title>\n<body>\n<h3>Top %d largest clones</h3>\n<p>\n<table><tr>\n", cnt, cnt);
			for (i = 0; i < g_top_cols; ++i) {
				fprintf(topF, "<th><font color=red>Weight</font></th><th>Clone</th>");
			}
			for (i = g_top_cols; --atP >= topP; ++i) {
				if (i == g_top_cols) {
					fprintf(topF, "</tr>\n<tr>");
					i = 0;
				}
				fprintf(topF, "<td align=right>%u</td><td><a href=%u.html>%u.html</td>", atP->weight, atP->html, atP->html);
			}
			fprintf(topF, "</tr>\n</table>\n<P>\nEnd of report\n</body>\n</html>\n");
			fclose(topF);
	}	}
}

static void
compute_lths(void)
{
	static Cfile	*fileP;
	static codeT	*code1P, *code2P;

	g_lth1     = end_code1P - start_code1P;
	g_lth2     = end_code2P - start_code2P;

	if (g_ignore_macros) {
		fileP  = 0;
		for (code1P = start_code1P; code1P < end_code1P; ++code1P) {
			if ((code2P = code1P->m_matchP)) {
				if (code1P->m_fileP         == code2P->m_fileP &&
					code1P->m_source_lineno == code2P->m_source_lineno) {
					if (fileP == code1P->m_fileP) {
						--g_lth1;
						--g_lth2;
						continue;
					}
					fileP = code1P->m_fileP;
					continue;
			}	}
			fileP = 0;
}	}	}

static int
permitted_clone(void)
{
	if ( (g_lth1 < g_min_match_sequence) || (g_lth2 < g_min_match_sequence)) {
		// Generally the clone is too short on one side or the other to be deemed worth reporting
		// Consider a clone long enough if it includes all of a function
		if ((start_code1P != function_code1P || end_code1P != function_end1P || g_lth1 < g_min_function_sequence) &&
			(start_code2P != function_code2P || end_code2P != function_end2P || g_lth2 < g_min_function_sequence)
		   ) {
			return(0);
	}	}
	return(1);
}

static int	g_check_forward_branches;	// True if we have some unchecked forward branches
static int	g_unknown_forward_branches;	// True if forward branches can't be fully checked

static int
same_branch(codeT *code1P, codeT *code2P)
{
	codeT	*to1P, *to2P, *matchP;

	if (code1P->m_jump < 0) {

		if (code2P->m_jump >= 0) {
			return g_mismatch_cost;
		}
		to2P = code2P + code2P->m_jump;
		if (to2P < start_code2P) {
			return g_mismatch_cost;
		}
		to1P = code1P + code1P->m_jump;
		if (to1P < start_code1P) {
			return g_mismatch_cost;
		}

		// A back branch matches if the things it points to match or
		// if the statement before it matches

		matchP = to1P->m_matchP;
		if (!matchP || matchP != to2P) {
			if (--to1P >= start_code1P && --to2P >= start_code2P) {
				matchP = to1P->m_matchP;
				if (matchP && matchP == to2P) {
					return g_same_cost;
			}	}
			return g_mismatch_cost;
		}
	} else if (code1P->m_jump > 0) {
		if (code2P->m_jump <= 0) {
			return g_mismatch_cost;
		}
		to2P = code2P + code2P->m_jump;
		if (end_code2P <= to2P) {
			return g_mismatch_cost;
		}
		to1P = code1P + code1P->m_jump;
		if (end_code1P <= to1P) {
			return g_mismatch_cost;
		}
		if (g_unknown_forward_branches) {
			++g_check_forward_branches;
		} else {
			matchP = to1P->m_matchP;
			if (!matchP || matchP != to2P) {
				if (--to1P >= code1P && --to2P >= code2P) {
					matchP = to1P->m_matchP;
					if (matchP && matchP == to2P) {
						return g_same_cost;
				}	}
				return g_mismatch_cost;
		}	}
	} else if (code2P->m_jump != 0) {
		// Self loops match same self loop
		return g_mismatch_cost;
	}
	return g_same_cost;
}

static int
same_code(codeT *code1P, codeT *code2P) 
{
	char	*assembler1P, *assembler2P;
	char	c;

	assembler1P = code1P->m_codeP;
	assembler2P = code2P->m_codeP;

	if (*assembler1P != *assembler2P) {
		return g_mismatch_cost;
	}
	if (code1P->m_jump != JUMP_UNDEFINED && code2P->m_jump != JUMP_UNDEFINED) {
		while (++assembler1P, ++assembler2P) {
			if ((c = *assembler1P) != *assembler2P) {
				return g_mismatch_cost;
			}
			switch (c) {
			case 0:
				return g_same_cost;
			case ' ':
				break;
			default:
				continue;
			}
			return same_branch(code1P, code2P);
	}	}

	if (strcmp(assembler1P, assembler2P)) {
		return g_mismatch_cost;
	}
	return g_same_cost;
}

static codeT *
detect_clone(void)
{
	static	matchesT	*matches_cacheP    = 0;
	static	matchesT	*matches_endP      = 0;

	static	codeT	*code1P, *code2P, *at_code1P, *at_code2P;
	static	codeT	*end1P, *end2P, *best_end1P, *best_end2P;
	static	codeT	*old_match1P, *old_match2P, *match1P, *match2P;
	static	int		best, distance, ret, weight;
	static	int		check_back_branches, recalculate_clone;

#ifdef DEBUG_DETECT
	static int debug_cnt = 0;

	fprintf(stderr, "%d Comparing %s(%d-%d) and %s(%d-%d)\n  %s\n  %s\n",  
		++debug_cnt,
		function1P->m_nameP,
		start_code1P - function_code1P,
		end_code1P - function_code1P,
		function2P->m_nameP,
		start_code2P - function_code2P, 
		end_code2P - function_code2P, 
		start_code1P->m_codeP, 
		start_code2P->m_codeP);
#endif
	weight = same_code(start_code1P, start_code2P);
	if (weight <= 0) {
		goto fail;
	}
	// Initially have to have a match
	best_end1P = code1P = start_code1P;
	best_end2P = code2P = start_code2P;
	code1P->m_matchP    = code2P;
	code2P->m_matchP    = code1P;

	g_check_forward_branches   = 0;
	g_unknown_forward_branches = 1;

	for (best = weight; ++code1P < end_code1P && ++code2P < end_code2P;) {
		for (distance = 0; weight >= 0; weight += g_mismatch_cost) {
			at_code1P = code1P;
			at_code2P = code2P + distance;
			if (at_code2P >= end_code2P) {
				at_code1P += (at_code2P - end_code2P + 1);
				if (at_code1P >= end_code1P) {
					break;
				}
				at_code2P  = end_code2P - 1;
			} 
			at_code2P->m_matchP = 0;
			while (at_code2P >= code2P && at_code1P < end_code1P) {
				ret = same_code(at_code1P, at_code2P);
				if (ret > 0) {
					goto found;
				}
				++at_code1P;
				--at_code2P;
			}
			at_code1P[-1].m_matchP = 0;
			++distance;
		}
		break;
found:	weight         += ret;
		end1P           = at_code1P;
		end2P           = at_code2P;
		end1P->m_matchP = end2P;
		end2P->m_matchP = end1P;

		if (g_use == 1) {
			for (code1P =  at_code1P; --code1P > start_code1P; ) {
				old_match2P = code1P->m_matchP;
				if (!old_match2P) {
					continue;
				}
				for (code2P =  match2P = old_match2P; ; ){
					code2P = code2P->m_hashP;
					if (code2P <= match2P || code2P >= at_code2P) {
						break;
					}
					if (same_code(code1P, code2P)) {
						// Advance match on the right side
						match2P = code2P;
				}	}
				match1P = old_match1P = code1P;
				for (;;) {
					code1P = code1P->m_hashP;
					if (code1P <= match1P || code1P >= at_code1P) {
						break;
					}
					if (same_code(code1P, match2P)) {			
						// Advance match on the left side
						match1P = code1P;
				}	}
				if (old_match1P == match1P && old_match2P == match2P) {
					break;
				}
				// Move the matched pair to the new match points
				old_match1P->m_matchP = 0;
				old_match2P->m_matchP = 0;
				match1P->m_matchP     = match2P;
				match2P->m_matchP     = match1P;
			
				// Now try to find new earlier new matches

				code1P = --match1P;
				code2P = --match2P;
				for (; !(match1P->m_matchP); --match1P);
				for (; !(match2P->m_matchP); --match2P);
				old_match1P = 0;
retry_fixup:	if (code1P == match1P || code2P == match2P) {
					break;
				}
				for (distance = 0; ; ++distance) {
					at_code1P = code1P;
					at_code2P = code2P - distance;
					if (at_code2P <= match2P) {
						at_code1P += (at_code2P - match2P - 1);
						if (at_code1P <= match1P) {
							break;
						}
						at_code2P  = match2P + 1;
					} 
					while (at_code2P <= code2P && at_code1P > match1P) {
						ret = same_code(at_code1P, at_code2P);
						if (ret > 0) {
							// fprintf(stderr, "Found extra match\n");
							weight += ret;
							old_match1P         = at_code1P;
							at_code1P->m_matchP = at_code2P;
							at_code2P->m_matchP = at_code1P;
							code1P              = at_code1P - 1;
							code2P              = at_code2P - 1;
							goto retry_fixup;
						}
						--at_code1P;
						++at_code2P;
				}	}
				if (!old_match1P) {
					break;
				}
				code1P = old_match1P;
		}	}
					
		if (weight >= best) {
			best        = weight;
			best_end1P  = end1P;
			best_end2P  = end2P;
		}
		code1P          = end1P;
		code2P          = end2P;
	}
	end_code1P = best_end1P+1;
	end_code2P = best_end2P+1;

	compute_lths();
	if (!permitted_clone()) {
		goto fail;
	}

	check_back_branches        = 0;
	recalculate_clone          = 0;
	g_unknown_forward_branches = 0;

	while (g_check_forward_branches || check_back_branches) {
		if (g_check_forward_branches) {
			g_check_forward_branches = 0;
			for (code1P = end_code1P; --code1P >= start_code1P; ) {
				if ((code2P = code1P->m_matchP)) {
					if (code1P->m_jump > 0) {
						if (!same_branch(code1P, code2P)) {
							code1P->m_matchP    = 0;
							code2P->m_matchP    = 0;
							if (code1P == start_code1P) {
								goto fail;
							}
							// Removing this match can invalidate a back branch
							// Can't invalidate a forward branch because scanning backwards
							check_back_branches = 1;
							recalculate_clone   = 1;
		}	}	}	}	}
		if (check_back_branches) {
			check_back_branches = 0;
			for (code1P = start_code1P; code1P < end_code1P; ++code1P) {
				if ((code2P = code1P->m_matchP)) {
					if (code1P->m_jump < 0) {
						if (!same_branch(code1P, code2P)) {
							code1P->m_matchP    = 0;
							code2P->m_matchP    = 0;
							// Removing this match can invalidate a forward branch
							// Can't invalidate a forward branch because scanning forwards
							g_check_forward_branches = 1;
	}	}	}	}	}	}

	if (recalculate_clone) {
		// Now having done the matching assuming at least some forward branches match recompute the clone size
		// on the basis of whether they do indeed match

		weight   = 0;
		code1P   = start_code1P;
		code2P   = start_code2P;
	
		for (; ; ) {
			if (end_code1P <= code1P) {
				if (end_code2P <= code2P) {
					break;
				}
			} else {
				if (!code1P->m_matchP) {
					weight += g_mismatch_cost;
					if (weight < 0) {
						break;
					}
					++code1P;
					continue;
			}	}
			if (code2P < end_code2P) {
				if (!code2P->m_matchP) {
					weight += g_mismatch_cost;
					if (weight < 0) {
						break;
					}
					++code2P;
					continue;
			}	}
			assert(code1P->m_matchP == code2P);
			assert(code2P->m_matchP == code1P);
			weight    += g_same_cost;
			best_end1P = ++code1P;
			best_end2P = ++code2P;
		}
		end_code1P = best_end1P;
		end_code2P = best_end2P;
		compute_lths();
		if (!permitted_clone()) {
			goto fail;
		}
	}
	
	report_clone();
	for (at_code1P = start_code1P; at_code1P < end_code1P; ++at_code1P) {
		if ((at_code2P = at_code1P->m_matchP)) {
			if (matches_cacheP == matches_endP) {
				matches_cacheP = (matchesT *) xmalloc(sizeof(matchesT) * 8192);
				matches_endP   = matches_cacheP + 8192;
			}
			matches_cacheP->m_nextP  = at_code1P->m_matchesP;
			matches_cacheP->m_codeP  = at_code2P;
			at_code1P->m_matchesP    = matches_cacheP++;
	}	}
	return(end_code2P);
fail:
	return(0);
}

void
detect(void)
{
	static	Cfunction		*function3P;
	static	matchesT		*matchesP;
	static	Cfile			*fileP;	
	static	codeT			*codes1P, *codes2P, *skipP, *lastP;

	for (source1P = Csource::g_headP; source1P; source1P = source1P->m_nextP) {
		if (g_verbose) {
			fprintf(stderr, "Matching %s\n", source1P->m_fileP->m_nameP);
		}
		codes1P = source1P->m_codesP;
		for (function1P = source1P->m_functionsP; function1P; function1P = function1P->m_nextP) {
			start_code1P   = function_code1P = codes1P + function1P->m_start_code;
			function_end1P = codes1P + function1P->m_end_code;
			for ( ; start_code1P < function_end1P; ++start_code1P) {
				if (g_match_lines) {
					if (start_code1P != function_code1P && (fileP = start_code1P->m_fileP)) {
						lastP = start_code1P - 1;
						if (lastP->m_fileP == fileP && lastP->m_source_lineno == start_code1P->m_source_lineno) {
							continue;
				}	}	}

				for (start_code2P = start_code1P->m_hashP; start_code2P; ) {
					function2P      = start_code2P->m_functionP;
					source2P        = function2P->m_sourceP;
					codes2P         = source2P->m_codesP;

					function_code2P = codes2P + function2P->m_start_code;
					function_end2P  = codes2P + function2P->m_end_code;
					if (g_match_lines && (g_use == 1)) {
						// If g_use is 0 we only hash lines that satisfy start of line
						if (start_code2P != function_code2P && (fileP = start_code2P->m_fileP)) {
							lastP = start_code2P - 1;
							if (lastP->m_fileP == fileP && lastP->m_source_lineno == start_code2P->m_source_lineno) {
								goto advance;
					}	}	}

					// Detect subsequence within larger sequences already matched
					for (matchesP = start_code1P->m_matchesP; matchesP; matchesP = matchesP->m_nextP) {
						if (matchesP->m_codeP == start_code2P) {
							goto advance;
					}	}

					if (function2P == function1P) {
						end_code2P = function_end1P;
						end_code1P = start_code2P;
					} else {
						end_code1P = function_end1P;
						end_code2P = function_end2P;
					}
					g_lth1 = end_code1P - start_code1P;
					g_lth2 = end_code2P - start_code2P;
					if (!permitted_clone()) {
						goto advance;
					}
				
					++g_comparisons;
					skipP         = detect_clone();
					start_code2P  = start_code2P->m_hashP;

					if (skipP) {
						for (; start_code2P; start_code2P = start_code2P->m_hashP) {
							function3P = start_code2P->m_functionP;
							if (function3P != function2P || start_code2P >= skipP) {
								break;
					}	}	}
					continue;
advance:			start_code2P = start_code2P->m_hashP;
}	}	}	}	}

